/*
GNU General Public License

GIFSave - GIF Image Writer
Copyright (C) 2000/2001  Howard A Kistler

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package com.hexidec.grafito;

import java.io.*;
import java.awt.*;

/** GIFSave
  *
  * NOTE: This class will be eventually superceded by my GIFCodec
  */

public class GIFSave
{
	short  imgWidth;
	short  imgHeight;
	int    numColors;
	byte[] pixelArray;
	byte[] colorArray;

//  Constructors ---------------------------------------------------------------------------->

	public GIFSave(int width, int height, int[] pixels) throws AWTException
	{
		imgWidth  = (short)width;
		imgHeight = (short)height;
		byte[][] rVals = new byte[imgWidth][imgHeight];
		byte[][] gVals = new byte[imgWidth][imgHeight];
		byte[][] bVals = new byte[imgWidth][imgHeight];
		for(int y = 0; y < imgHeight; y++)
		{
			for(int x = 0; x < imgWidth; x++)
			{
				rVals[x][y] = (byte)((pixels[(y * imgWidth) + x] >> 16) & 0xFF);
				gVals[x][y] = (byte)((pixels[(y * imgWidth) + x] >> 8)  & 0xFF);
				bVals[x][y] = (byte)((pixels[(y * imgWidth) + x])       & 0xFF);  
			}
		}
		toIndexedColor(rVals, gVals, bVals);
	}

//  Class Methods --------------------------------------------------------------------------->

	private void toIndexedColor(byte[][] rVals, byte[][] gVals, byte[][] bVals) throws AWTException
	{
		pixelArray = new byte[imgWidth * imgHeight];
		colorArray = new byte[256 * 3];
		int colornum = 0;
		for(int y = 0; y < imgHeight; y++)
		{
			for(int x = 0; x < imgWidth; x++)
			{
				int search;
				for(search = 0; search < colornum; search++)
				{
					if(
						(colorArray[search * 3]     == rVals[x][y]) &&
						(colorArray[search * 3 + 1] == gVals[x][y]) &&
						(colorArray[search * 3 + 2] == bVals[x][y])
						)
					{
						break;
					}
				}
				if(search > 255)
				{
					throw new AWTException("Too many colors");
				}
				pixelArray[(y * imgWidth) + x] = (byte)search;
				if(search == colornum)
				{
					colorArray[search * 3]     = rVals[x][y];
					colorArray[search * 3 + 1] = gVals[x][y];
					colorArray[search * 3 + 2] = bVals[x][y];
					colornum++;
				}
			}
		}
		numColors = 1 << getBitsNeeded(colornum);
		byte copy[] = new byte[numColors * 3];
		System.arraycopy(colorArray, 0, copy, 0, numColors * 3);
		colorArray = copy;
	}

	protected void writeImageToFile(OutputStream output) throws IOException
	{
		byte TypeByte1 = 'G'; // Writes GIFs of type GIF89a
		byte TypeByte2 = 'I';
		byte TypeByte3 = 'F';
		byte TypeByte4 = '8';
		byte TypeByte5 = '9';
		byte TypeByte6 = 'a';
		byte BitmapWidthByte1  = (byte)( imgWidth        & 0xFF);
		byte BitmapWidthByte2  = (byte)((imgWidth >> 8)  & 0xFF);
		byte BitmapHeightByte1 = (byte)( imgHeight       & 0xFF);
		byte BitmapHeightByte2 = (byte)((imgHeight >> 8) & 0xFF);
		byte MaskingByte = (byte)(-16);
		byte colorMask   = (byte)(getBitsNeeded(numColors) - 1);
		MaskingByte      = (byte)(MaskingByte | (colorMask & 7));
		byte BackgroundColorIndexByte = (byte)0;
		byte AspectRatioByte = (byte)0;

		/* Build and write the file header */
		byte[] BitmapFileHeader = new byte[]
		{
			TypeByte1,         TypeByte2,         TypeByte3, TypeByte4, TypeByte5, TypeByte6,
			BitmapWidthByte1,  BitmapWidthByte2,
			BitmapHeightByte1, BitmapHeightByte2,
			MaskingByte,
			BackgroundColorIndexByte,
			AspectRatioByte
		};
		output.write(BitmapFileHeader);

		/* Write the color map of color values and color indices */
		output.write(colorArray, 0, colorArray.length);

		/* Build and write the image header */
		byte DataDelimiterByte  = ',';
		byte HorizontalPosByte1 = (byte)0;
		byte HorizontalPosByte2 = (byte)0;
		byte VerticalPosByte1   = (byte)0;
		byte VerticalPosByte2   = (byte)0;
		byte ImageMaskingByte   = (byte)0;

		byte[] BitmapImageHeader = new byte[]
		{
			DataDelimiterByte,
			HorizontalPosByte1, HorizontalPosByte2,
			VerticalPosByte1,   VerticalPosByte2,
			BitmapWidthByte1,   BitmapWidthByte2,
			BitmapHeightByte1,  BitmapHeightByte2,
			ImageMaskingByte
		};
		output.write(BitmapImageHeader);

		/* Write the body size (based on number of colors in use) */
		byte bodySize = getBitsNeeded(numColors);
		if(bodySize == 1) { bodySize++; }
		output.write(bodySize);

		/* Compress the body into GIF LZW format */
///////////////////////<
		LZWCompressor.LZWCompress(output, bodySize, pixelArray);
///////////////////////>

		/* Build and write the image footer */
		DataDelimiterByte  = ';';
		HorizontalPosByte1 = (byte)0;
		HorizontalPosByte2 = (byte)0;
		VerticalPosByte1   = (byte)0;
		VerticalPosByte2   = (byte)0;
		ImageMaskingByte   = (byte)0;

		byte[] BitmapImageFooter = new byte[]
		{
			DataDelimiterByte,
			HorizontalPosByte1, HorizontalPosByte2,
			VerticalPosByte1,   VerticalPosByte2,
			BitmapWidthByte1,   BitmapWidthByte2,
			BitmapHeightByte1,  BitmapHeightByte2,
			ImageMaskingByte
		};
		output.write(BitmapImageFooter);

		output.flush();
		output.close();
	}

//  Convenience Methods --------------------------------------------------------------------->

	private byte getBitsNeeded(int n)
	{
		byte ret = 1;
		if(n-- == 0)
		{
			return 0;
		}
		while ((n >>= 1) != 0)
		{
			ret++;
		}
		return ret;
	}    

}

//  Ancillary Classes ----------------------------------------------------------------------->

	class BitFile
	{
		OutputStream output_;
		byte buffer_[];
		int index_, bitsLeft_;

		public BitFile(OutputStream output)
		{
			output_ = output;
			buffer_ = new byte[256];
			index_ = 0;
			bitsLeft_ = 8;
		}

		public void flush() throws IOException
		{
			int numBytes = index_ + (bitsLeft_ == 8 ? 0 : 1);
			if(numBytes > 0)
			{
				output_.write(numBytes);
				output_.write(buffer_, 0, numBytes);
				buffer_[0] = 0;
				index_ = 0;
				bitsLeft_ = 8;
			}
		}

		public void writeBits(int bits, int numbits) throws IOException
		{
			int bitsWritten = 0;
			int numBytes = 255;
			do {
				if((index_ == 254 && bitsLeft_ == 0) || index_ > 254)
				{
					output_.write(numBytes);
					output_.write(buffer_, 0, numBytes);
					buffer_[0] = 0;
					index_ = 0;
					bitsLeft_ = 8;
				}

				if(numbits <= bitsLeft_)
				{
					buffer_[index_] |= (bits & ((1 << numbits) - 1)) << (8 - bitsLeft_);
					bitsWritten += numbits;
					bitsLeft_ -= numbits;
					numbits = 0;
				}
				else
				{
					buffer_[index_] |= (bits & ((1 << bitsLeft_) - 1)) << (8 - bitsLeft_);
					bitsWritten += bitsLeft_;
					bits >>= bitsLeft_;
					numbits -= bitsLeft_;
					buffer_[++index_] = 0;
					bitsLeft_ = 8;
				}
			} while (numbits != 0);
		}
	}

	class LZWStringTable
	{
		private final static int RES_CODES = 2;
		private final static short HASH_FREE = (short)0xFFFF;
		private final static short NEXT_FIRST = (short)0xFFFF;
		private final static int MAXBITS = 12;
		private final static int MAXSTR = (1 << MAXBITS);
		private final static short HASHSIZE = 9973;
		private final static short HASHSTEP = 2039;

		byte strChr_[];
		short strNxt_[];
		short strHsh_[];
		short numStrings_;

		public LZWStringTable()
		{
			strChr_ = new byte[MAXSTR];
			strNxt_ = new short[MAXSTR];
			strHsh_ = new short[HASHSIZE];    
		}

		public int addCharString(short index, byte b)
		{
			int hshidx;

			if(numStrings_ >= MAXSTR)
			{
				return 0xFFFF;
			}

			hshidx = Hash(index, b);
			while (strHsh_[hshidx] != HASH_FREE)
			{
				hshidx = (hshidx + HASHSTEP) % HASHSIZE;
			}

			strHsh_[hshidx] = numStrings_;
			strChr_[numStrings_] = b;
			strNxt_[numStrings_] = (index != HASH_FREE) ? index : NEXT_FIRST;

			return numStrings_++;
		}

		public short findCharString(short index, byte b)
		{
			int hshidx, nxtidx;

			if(index == HASH_FREE)
			{
				return b;
			}

			hshidx = Hash(index, b);
			while ((nxtidx = strHsh_[hshidx]) != HASH_FREE)
			{
				if(strNxt_[nxtidx] == index && strChr_[nxtidx] == b)
				{
					return (short)nxtidx;
				}
				hshidx = (hshidx + HASHSTEP) % HASHSIZE;
			}

			return (short)0xFFFF;
		}

		public void ClearTable(int codesize)
		{
			numStrings_ = 0;

			for(int q = 0; q < HASHSIZE; q++)
			{
				strHsh_[q] = HASH_FREE;
			}

			int w = (1 << codesize) + RES_CODES;
			for(int q = 0; q < w; q++)
			{
				addCharString((short)0xFFFF, (byte)q);
			}
		}

		static public int Hash(short index, byte lastbyte)
		{
			return ((int)((short)(lastbyte << 8) ^ index) & 0xFFFF) % HASHSIZE;
		}
	}

	class LZWCompressor
	{

		public static void LZWCompress(OutputStream output, int codesize, byte toCompress[]) throws IOException
		{
			byte c;
			short index;
			int clearcode, endofinfo, numbits, limit, errcode;
			short prefix = (short)0xFFFF;

			BitFile bitFile = new BitFile(output);
			LZWStringTable strings = new LZWStringTable();

			clearcode = 1 << codesize;
			endofinfo = clearcode + 1;

			numbits = codesize + 1;
			limit = (1 << numbits) - 1;

			strings.ClearTable(codesize);
			bitFile.writeBits(clearcode, numbits);

			for(int i = 0; i < toCompress.length; i++)
			{
				c = toCompress[i];
				if((index = strings.findCharString(prefix, c)) != -1)
				{
					prefix = index;
				}
				else
				{
					bitFile.writeBits(prefix, numbits);
					if(strings.addCharString(prefix, c) > limit)
					{
						if(++numbits > 12)
						{
							bitFile.writeBits(clearcode, numbits - 1);
							strings.ClearTable(codesize);
							numbits = codesize + 1;
						}
						limit = (1 << numbits) - 1;
					}
					prefix = (short)((short)c & 0xFF);
				}
			}

			if(prefix != -1)
			{
				bitFile.writeBits(prefix, numbits);
			}

			bitFile.writeBits(endofinfo, numbits);
			bitFile.flush();
		}
	}


